"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from moderngl_window import resources
from moderngl_window.meta import ProgramDescription

import functools

def _load(name, defines={}):
    return resources.programs.load(ProgramDescription(path=name, defines=defines))


@functools.lru_cache()
def get_lit_program(vs, smooth_shading, texture, face_color, instanced=0):
    defines = {
        'SMOOTH_SHADING': smooth_shading,
        'TEXTURE': texture,
        'FACE_COLOR': face_color,
        'INSTANCED': instanced,
    }
    return resources.programs.load(ProgramDescription(vertex_shader=vs,
                                                      geometry_shader="lit_with_edges.glsl",
                                                      fragment_shader="lit_with_edges.glsl",
                                                      defines=defines))


def get_smooth_lit_with_edges_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=1, texture=0, face_color=0, instanced=instanced)


def get_smooth_lit_with_edges_face_color_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=1, texture=0, face_color=1, instanced=instanced)


def get_flat_lit_with_edges_face_color_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=0, texture=0, face_color=1, instanced=instanced)


def get_flat_lit_with_edges_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=0, texture=0, face_color=0, instanced=instanced)


def get_flat_lit_with_edges_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=0, texture=0, face_color=0, instanced=instanced)


def get_smooth_lit_texturized_program(vs, instanced=0):
    return get_lit_program(vs, smooth_shading=1, texture=1, face_color=0, instanced=instanced)


def get_sphere_instanced_program():
    return get_smooth_lit_with_edges_program("sphere_instanced.vs.glsl")


def get_lines_instanced_program():
    return get_smooth_lit_with_edges_program("lines_instanced.vs.glsl")


@functools.lru_cache()
def get_outline_program(vs_path, instanced=0):
    defines = {
        'INSTANCED': instanced
    }
    return resources.programs.load(ProgramDescription(vertex_shader=vs_path,
                                                      fragment_shader="outline/outline_prepare.fs.glsl",
                                                      defines=defines))


@functools.lru_cache()
def get_fragmap_program(vs_path, instanced=0):
    defines = {
        'INSTANCED': instanced
    }
    return resources.programs.load(ProgramDescription(vertex_shader=vs_path,
                                                      fragment_shader="fragment_picking/frag_map.fs.glsl",
                                                      defines=defines))


@functools.lru_cache()
def get_depth_only_program(vs_path, instanced=0):
    defines = {
        'INSTANCED': instanced
    }
    return resources.programs.load(ProgramDescription(vertex_shader=vs_path,
                                                      fragment_shader="shadow_mapping/depth_only.fs.glsl",
                                                      defines=defines))


@functools.lru_cache()
def get_simple_unlit_program():
    return _load('simple_unlit.glsl')


@functools.lru_cache()
def get_cylinder_program():
    return _load('cylinder.glsl')


@functools.lru_cache()
def get_screen_texture_program():
    return _load('screen_texture.glsl')


@functools.lru_cache()
def get_chessboard_program():
    return _load('chessboard.glsl')


def clear_shader_cache():
    """Clear all cached shaders."""
    funcs =  [
        get_lit_program,
        get_outline_program,
        get_fragmap_program,
        get_depth_only_program,
        get_simple_unlit_program,
        get_cylinder_program,
        get_screen_texture_program,
        get_chessboard_program,
    ]
    for f in funcs:
        f.cache_clear()




"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import torch

from aitviewer.configuration import CONFIG as C
from aitviewer.models.star import STARLayer
from aitviewer.renderables.smpl import SMPLSequence
from aitviewer.utils import to_numpy as c2c


class STARSequence(SMPLSequence):
    """
    Represents a temporal sequence of SMPL poses using the STAR model.
    """

    def __init__(self,
                 poses_body,
                 smpl_layer,
                 poses_root,
                 betas=None,
                 trans=None,
                 device=C.device,
                 include_root=True,
                 normalize_root=False,
                 is_rigged=True,
                 show_joint_angles=False,
                 z_up=False,
                 post_fk_func=None,
                 **kwargs):

        super(STARSequence, self).__init__(poses_body, smpl_layer, poses_root, betas, trans, device=device,
                                           include_root=include_root, normalize_root=normalize_root,
                                           is_rigged=is_rigged, show_joint_angles=show_joint_angles, z_up=z_up,
                                           post_fk_func=post_fk_func, **kwargs)

    def fk(self, current_frame_only=False):
        """Get joints and/or vertices from the poses."""
        if current_frame_only:
            # Use current frame data.
            if self._edit_mode:
                poses_root = self._edit_pose[:3][None, :]
                poses_body = self._edit_pose[3:][None, :]
            else:
                poses_body = self.poses_body[self.current_frame_id][None, :]
                poses_root = self.poses_root[self.current_frame_id][None, :]

            trans = self.trans[self.current_frame_id][None, :]

            if self.betas.shape[0] == self.n_frames:
                betas = self.betas[self.current_frame_id][None, :]
            else:
                betas = self.betas
        else:
            # Use the whole sequence.
            if self._edit_mode:
                poses_root = self.poses_root.clone()
                poses_body = self.poses_body.clone()
                poses_root[self.current_frame_id] = self._edit_pose[:3]
                poses_body[self.current_frame_id] = self._edit_pose[3:]
            else:
                poses_body = self.poses_body
                poses_root = self.poses_root
            trans = self.trans
            betas = self.betas

        verts, joints = self.smpl_layer(poses_root=poses_root,
                                        poses_body=poses_body,
                                        betas=betas,
                                        trans=trans,
                                        normalize_root=self._normalize_root
                                        )

        skeleton = self.smpl_layer.skeletons()['body'].T
        faces = self.smpl_layer.faces
        joints = joints[:, :skeleton.shape[0]]

        if current_frame_only:
            return c2c(verts)[0], c2c(joints)[0], c2c(faces), c2c(skeleton)
        else:
            return c2c(verts), c2c(joints), c2c(faces), c2c(skeleton)

    @classmethod
    def from_amass(cls,
                   npz_data_path,
                   start_frame=None,
                   end_frame=None,
                   sub_frames=None,
                   log=True,
                   fps_out=None,
                   load_betas=False,
                   z_up=True,
                   **kwargs):
        """Load a sequence downloaded from the AMASS website."""

        # User SMPL sequence loader and re-parse data
        seq = super().from_amass(npz_data_path, start_frame=start_frame, end_frame=end_frame, log=log, fps_out=fps_out,
                                 **kwargs)

        # STAR has no hands, but includes wrists
        poses_body = torch.cat((seq.poses_body, seq.poses_left_hand[:, :3], seq.poses_right_hand[:, :3]), dim=-1)
        poses_root = seq.poses_root
        trans = seq.trans
        betas = None

        if load_betas:
            print("WARNING: Loading betas from AMASS into STAR requires an optimization procedure. " +
                  "See https://github.com/ahmedosman/STAR/tree/master/convertors")
            betas = seq.betas

        if sub_frames is not None:
            poses_root = poses_root[sub_frames]
            poses_body = poses_body[sub_frames]
            trans = trans[sub_frames]

        return cls(poses_body=poses_body,
                   smpl_layer=STARLayer(device=C.device),
                   poses_root=poses_root,
                   betas=betas,
                   trans=trans,
                   include_root=seq._include_root,
                   is_rigged=seq._is_rigged,
                   z_up=z_up,
                   color=seq.color,
                   **kwargs)

    @classmethod
    def from_3dpw(cls):
        raise ValueError('STAR does not support loading from 3DPW.')

    @classmethod
    def t_pose(cls, model=None, betas=None, frames=1, **kwargs):
        """Creates a SMPL sequence whose single frame is a SMPL mesh in T-Pose."""

        if model is None:
            model = STARLayer(device=C.device)

        poses_body = np.zeros([frames, model.n_joints_body * 3])
        poses_root = np.zeros([frames, 3])
        return cls(poses_body=poses_body, smpl_layer=model, poses_root=poses_root, betas=betas, **kwargs)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import os
import re

from collections import OrderedDict
from aitviewer.scene.camera import OpenCVCamera
from aitviewer.scene.node import Node
from aitviewer.renderables.billboard import Billboard


class MultiViewSystem(Node):
    """A multi view camera system which can be used to visualize cameras and images"""

    def __init__(self, camera_info_path, camera_images_path, cols, rows, viewer, **kwargs):
        """
        Load a MultiViewSystem from a file with camera information for M cameras
        and the path to a directory with M subdirectories, containing N images each.

        :param camera_info_path: path to a camera info .npz file with the following file entries
            'ids':         A np array of camera ids (M)
            'intrinsics':  A np array of camera intrinsics in the format used by OpenCV (M, 3, 3)
            'extrinsics':  A np array of camera extrinsics in the format used by OpenCV (M, 3, 4)
            'dist_coeffs': A np array of camera distortion coefficients in the format used by OpenCV (M, 5)
        :param camera_images_path: path to a directory with M subdirectories, one for each camera
            with the camera id as name. Each directory contains N images for the respective camera
            with a filename ending with the frame number.
            Example:
            camera_images_path/1/image001.png
            camera_images_path/1/image002.png
            camera_images_path/2/image001.png
            camera_images_path/2/image002.png
        :param cols: width  of the image in pixels, matching the size of the image expected by the intrinsics matrix
        :param rows: height of the image in pixels, matching the size of the image expected by the intrinsics matrix
        :param viewer: the viewer, used for changing the view to one of the cameras'
        """
        # Load camera information.
        camera_info = np.load(camera_info_path)

        # Compute max number of images per camera and use it as number of frames for this node.
        camera_n_frames = []
        for id in camera_info['ids']:
            path = os.path.join(camera_images_path, str(id))
            if os.path.isdir(path):
                camera_n_frames.append(len(os.listdir(path)))

        n_frames = max(camera_n_frames) if camera_n_frames else 1

        # Initialize node.
        super(MultiViewSystem, self).__init__(n_frames=n_frames, **kwargs)

        # Compute position and orientation of each camera.
        positions = []
        self.cameras = []
        for i in range(len(camera_info['ids'])):
            intrinsics = camera_info['intrinsics'][i]
            extrinsics = camera_info['extrinsics'][i]
            dist_coeffs = camera_info['dist_coeffs'][i]

            camera = OpenCVCamera(intrinsics, extrinsics, cols, rows, dist_coeffs=dist_coeffs, viewer=viewer,
                                  is_selectable=False)
            self.add(camera, show_in_hierarchy=False)
            self.cameras.append(camera)

            positions.append(camera.position)

        # Compute maximum distance from a camera to the center of all cameras and use it
        # to compute the distance at which to show the billboards.
        positions = np.array(positions)
        camera_center = np.mean(positions, 0)
        max_dist = np.max(np.apply_along_axis(lambda x: np.linalg.norm(x - camera_center), 1, positions))
        self.billboard_distance = max_dist * 2
        self.camera_positions = positions

        self.viewer = viewer
        self.camera_info = camera_info
        self.camera_images_path = camera_images_path
        self.cols = cols
        self.rows = rows

        # Maps from active camera index to its billboard or None if billboards are disabled.
        self.active_cameras = OrderedDict()

        self._billboards_enabled = False
        self._frustums_enabled = False
        self._cameras_enabled = True

        self.selected_camera_index = None

    def _create_billboard_for_camera(self, camera_index):
        """Helper function to create a billboard for the camera at the given index."""
        # Look for images for the given camera.
        camera_path = os.path.join(self.camera_images_path, str(self.camera_info['ids'][camera_index]))
        if not os.path.isdir(camera_path):
            print(f"Camera directory not found at {camera_path}")
            return

        # Sort images by the frame number in the filename.
        files = os.listdir(camera_path)
        regex = re.compile(r"(\d*)$")

        def sort_key(x):
            name = os.path.splitext(x)[0]
            return int(regex.search(name).group(0))

        paths = [os.path.join(camera_path, f) for f in sorted(files, key=sort_key)]
        if not paths:
            print(f"Camera images not found at {camera_path}")
            return

        # Create a new billboard for the currently active camera.
        billboard = Billboard.from_camera_and_distance(self.cameras[camera_index], self.billboard_distance, self.cols,
                                                       self.rows, paths)

        # Set the current frame index if we have an image for it.
        if self.current_frame_id < len(paths):
            billboard.current_frame_id = self.current_frame_id

        return billboard

    def activate_camera(self, index):
        """Activates the camera at the given index, showing its frustum and billboard if enabled."""
        if index not in self.active_cameras:
            # Change camera color.
            camera = self.cameras[index]
            camera.active = True
            # Create a billboard if billboards are enabled.
            billboard = None
            if self._billboards_enabled:
                billboard = self._create_billboard_for_camera(index)
                self.add(billboard, show_in_hierarchy=False)
            self.active_cameras[index] = billboard
            # Show the camera furstum if frustums are enabled.
            if self._frustums_enabled:
                camera.show_frustum(self.cols, self.rows, self.billboard_distance)

    def deactivate_camera(self, index):
        """Deactivates the camera at the given index, hiding its frustum and billboard if enabled."""
        if index in self.active_cameras:
            # Change camera color.
            camera = self.cameras[index]
            camera.active = False
            # Hide frustum.
            camera.hide_frustum()
            # Remove billboard if it exists.
            billboard = self.active_cameras[index]
            if billboard:
                self.remove(billboard)
            # Remove camera from active cameras
            del self.active_cameras[index]

    def view_from_camera(self, index):
        """
        View from the camera with the given index deactivating all other cameras,
        this also enables the billboard and disables the frustums.
        """
        self.billboards_enabled = True
        self.frustums_enabled = False
        active_cameras = list(self.active_cameras.keys())
        for i in active_cameras:
            if i != index:
                self.deactivate_camera(i)
        self.activate_camera(index)
        self.cameras[index].view_from_camera()

    @property
    def frustums_enabled(self):
        """Returns True if the frustums are enabled and False otherwise."""
        return self._frustums_enabled

    @frustums_enabled.setter
    def frustums_enabled(self, enabled):
        """Setting this to True shows the frustums of active cameras."""
        if enabled == self._frustums_enabled:
            return
        self._frustums_enabled = enabled

        # Update all frustums of all active cameras.
        for i in self.active_cameras.keys():
            if enabled:
                self.cameras[i].show_frustum(self.cols, self.rows, self.billboard_distance)
            else:
                self.cameras[i].hide_frustum()

    @property
    def billboards_enabled(self):
        """Returns True if the billboards are enabled and False otherwise."""
        return self._billboards_enabled

    @billboards_enabled.setter
    def billboards_enabled(self, enabled):
        """Setting this to True shows the billobards of active cameras."""
        if enabled == self._billboards_enabled:
            return
        self._billboards_enabled = enabled

        # Update billboards  of all active cameras.
        for index, billboard in self.active_cameras.items():
            if enabled:
                billboard = self._create_billboard_for_camera(index)
                self.add(billboard, show_in_hierarchy=False)
                self.active_cameras[index] = billboard
            else:
                # Billboard can be None here if the camera images were not found.
                if billboard is not None:
                    self.remove(billboard)
                self.active_cameras[index] = None

    @property
    def cameras_enabled(self):
        """Returns True if the cameras are enabled and False otherwise."""
        return self._cameras_enabled

    @cameras_enabled.setter
    def cameras_enabled(self, enabled):
        """Setting this to True shows the cameras as meshes in the scene."""
        if enabled == self._cameras_enabled:
            return
        self._cameras_enabled = enabled

        # Update all cameras.
        for c in self.cameras:
            c.enabled = enabled

    @property
    def bounds(self):
        return self.get_bounds(self.camera_positions)

    @property
    def current_bounds(self):
        return self.bounds

    def _gui_checkboxes(self, imgui):
        _, self.cameras_enabled = imgui.checkbox("Show cameras", self.cameras_enabled)
        _, self.billboards_enabled = imgui.checkbox("Show billboard", self.billboards_enabled)
        _, self.frustums_enabled = imgui.checkbox("Show frustum", self.frustums_enabled)

    def gui(self, imgui):
        self._gui_checkboxes(imgui)
        imgui.spacing()

        active_index = -1
        if len(self.active_cameras) > 0:
            active_index = next(reversed(self.active_cameras.keys()))
        u_selected, active_index = imgui.combo("ID", active_index, [str(id) for id in self.camera_info['ids'].tolist()])
        if u_selected:
            self.view_from_camera(active_index)

        imgui.text("Active cameras:")
        active_cameras = list(self.active_cameras.keys())
        for i in active_cameras:
            if imgui.button(f"{self.camera_info['ids'][i]}", width=50):
                self.view_from_camera(i)
            imgui.same_line(spacing=10)
            if imgui.button(f"x##{i}"):
                self.deactivate_camera(i)

    def gui_context_menu(self, imgui):
        if self.selected_camera_index is None:
            return

        imgui.text(f"Camera {self.camera_info['ids'][self.selected_camera_index]}")
        imgui.separator()

        u, s = imgui.menu_item(f"Activate camera", selected=self.selected_camera_index in self.active_cameras)
        if u:
            if s:
                self.activate_camera(self.selected_camera_index)
            else:
                self.deactivate_camera(self.selected_camera_index)

        _, v = imgui.menu_item(f"View from camera")
        if v:
            self.view_from_camera(self.selected_camera_index)

        imgui.spacing()
        imgui.spacing()
        imgui.text(f"{self.name}")
        imgui.separator()

        self._gui_checkboxes(imgui)

    def on_selection(self, node, instance_id, tri_id):
        # Find which camera is selected.
        for idx, c in enumerate(self.cameras):
            # A camera is selected if the selected node is one of its children.
            if node in c.nodes:
                self.selected_camera_index = idx
                break

    def render_outline(self, *args, **kwargs):
        # Render outline of the currently selected camera.
        if self.selected_camera_index is not None:
            self.cameras[self.selected_camera_index].render_outline(*args, **kwargs)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np

from aitviewer.configuration import CONFIG as C
from aitviewer.scene.material import Material
from aitviewer.utils.so3 import euler2rot_numpy, rot2euler_numpy
from functools import lru_cache


class Node(object):
    """Interface for nodes."""

    def __init__(self,
                 name=None,
                 icon=None,
                 position=None,
                 rotation=None,
                 scale=1.0,
                 color=(0.5, 0.5, 0.5, 1.0),
                 material=None,
                 is_selectable=True,
                 gui_affine=True,
                 gui_material=True,
                 enabled_frames=None,
                 n_frames=1):
        """
        :param name: Name of the node
        :param icon: Custom Node Icon using custom Icon font
        :param position: Starting position in the format (X,Y,Z) or np array of positions with shape (F, 3)
        :param rotation: Starting rotation in rotation matrix representation (3,3) or np array of rotations with shape (F, 3, 3)
        :param scale: Starting scale (scalar) or np array of scale values with shape (F)
        :param color: (R,G,B,A) 0-1 formatted color value.
        :param material: Object material properties. The color specified in the material will override node color
        :param is_selectable: If True the node is selectable when clicked on, otherwise the parent node will be selected.
        :param gui_affine: If True the node will have transform controls (position, rotation, scale) in the GUI.
        :param gui_material: If True the node will have material controls in the GUI.
        :param enabled_frames: Numpy array of boolean values, the object will be enabled only in frames where the value is True,
         the number of ones in the mask must match the number of frames of the object.
        :param n_frames: How many frames this renderable has.
        """
        # Transform & Animation
        position = np.zeros(3, dtype=np.float32) if position is None else np.array(position, dtype=np.float32)
        rotation = np.eye(3, dtype=np.float32) if rotation is None else np.array(rotation, dtype=np.float32)

        self._positions = position if len(position.shape) != 1 else position[np.newaxis]
        self._rotations = rotation if len(rotation.shape) != 2 else rotation[np.newaxis]
        self._scales = (scale if isinstance(scale, np.ndarray) else np.array([scale])).astype(np.float32)

        n_positions = self._positions.shape[0]
        n_rotations = self._rotations.shape[0]
        n_scales = self._scales.shape[0]

        if n_frames > 1:
            assert n_positions == 1 or n_frames == n_positions, (f"Number of position frames"
                f" ({n_positions}) must be 1 or match number of Node frames {n_frames}")
            assert n_rotations == 1 or n_frames == n_rotations, (f"Number of rotations frames"
                f" ({n_rotations}) must be 1 or match number of Node frames {n_frames}")
            assert n_scales == 1 or n_frames == n_scales, (f"Number of scales frames"
                f" ({n_scales}) must be 1 or match number of Node frames {n_frames}")
        else:
            n_frames = max(n_positions, n_rotations, n_scales)
            assert ((n_positions == 1 or n_positions == n_frames) and
                    (n_rotations == 1 or n_rotations == n_frames) and
                    (n_scales == 1 or n_scales == n_frames)), (f"Number of position"
                        f"({n_positions}), rotation ({n_rotations}) and scale ({n_scales})"
                        "frames must be 1 or match.")

        # Frames
        self._n_frames = n_frames
        self._current_frame_id = 0
        self.model_matrix = self.get_local_transform()
        self._enabled_frames = enabled_frames
        if self._enabled_frames is not None:
            assert np.count_nonzero(self._enabled_frames) == n_frames, (f"Number of non-zero elements in enabled_frames"
                f" ({np.count_nonzero(self._enabled_frames)}) must match number of frames in sequence ({n_frames})")
            # Create an array that maps from the true frame id (counting also disabled frames) to the index of the
            # first existing frame in the sequence.
            self._enabled_frame_id = np.cumsum(self._enabled_frames) - 1

            # Stores the true frame id (counting also disabled frames) we use this to allow going
            # through both enabled and disabled frames from the GUI.
            self._internal_frame_id = 0

        # Material
        self.material = Material(color=color) if material is None else material

        # Renderable Attributes
        self.is_renderable = False
        self.backface_culling = True
        self.backface_fragmap = False
        self.draw_outline = False

        # Flags to enable rendering passes
        self.cast_shadow = False
        self.depth_prepass = False
        self.fragmap = False
        self.outline = False

        # Programs for render passes. Subclasses are responsible for setting these.
        self.depth_only_program = None # Required for depth_prepass and cast_shadow passes
        self.fragmap_program = None    # Required for framap pass
        self.outline_program = None    # Required for outline pass

        # GUI
        self.name = name if name is not None else type(self).__name__
        self.uid = C.next_gui_id()
        self.unique_name = self.name + "{}".format(self.uid)
        self.icon = icon if icon is not None else '\u0082'
        self._enabled = True
        self._expanded = False
        self.gui_controls = {
            'affine': {'fn': self.gui_affine, 'icon': '\u009b', 'is_visible': gui_affine},
            'material': {'fn': self.gui_material, 'icon': '\u0088', 'is_visible': gui_material},
            'animation': {'fn': self.gui_animation, 'icon': '\u0098', 'is_visible': (lambda: self._n_frames > 1)()},
            'io': {'fn': self.gui_io, 'icon': '\u009a', 'is_visible': (lambda: self.gui_io.__func__ is not Node.gui_io)()}
        }
        self.gui_modes = {
            'view': {'title': ' View', 'fn': self.gui_mode_view, 'icon': '\u0099'}
        }
        self._selected_mode = 'view'
        self._show_in_hierarchy = True
        self.is_selectable = is_selectable

        self.nodes = []
        self.parent = None

    # Selected Mode
    @property
    def selected_mode(self):
        return self._selected_mode

    @selected_mode.setter
    def selected_mode(self, selected_mode):
        self._selected_mode = selected_mode

    # Transform
    @property
    def position(self):
        idx = self.current_frame_id if self._positions.shape[0] > 1 else 0
        return self._positions[idx]

    @position.setter
    def position(self, position):
        idx = self.current_frame_id if self._positions.shape[0] > 1 else 0
        self._positions[idx] = np.array(position, dtype=np.float32).copy()
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @property
    def positions(self):
        return self._positions

    @positions.setter
    def positions(self, positions):
        self._positions = positions
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @property
    def rotation(self):
        idx = self.current_frame_id if self._rotations.shape[0] > 1 else 0
        return self._rotations[idx]

    @rotation.setter
    def rotation(self, rotation):
        idx = self.current_frame_id if self._rotations.shape[0] > 1 else 0
        self._rotations[idx] = rotation
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @property
    def rotations(self):
        return self._rotations

    @rotations.setter
    def rotations(self, rotations):
        self._rotations = rotations
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @property
    def scale(self):
        idx = self.current_frame_id if self._scales.shape[0] > 1 else 0
        return self._scales[idx]

    @scale.setter
    def scale(self, scale):
        idx = self.current_frame_id if self._scales.shape[0] > 1 else 0
        self._scales[idx] = scale
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @property
    def scales(self):
        return self._scales

    @scales.setter
    def scales(self, scales):
        self._scales = scales
        self.update_transform(None if self.parent is None else self.parent.model_matrix)

    @staticmethod
    @lru_cache()
    def _compute_transform(pos, rot, scale):
        rotation = np.eye(4)
        rotation[:3, :3] = np.array(rot)

        trans = np.eye(4)
        trans[:3, 3] = np.array(pos)

        scale = np.diag([scale, scale, scale, 1])

        return (trans @ rotation @ scale).astype('f4')

    def get_local_transform(self):
        """Construct local transform as a 4x4 matrix from this node's position, orientation and scale."""
        return self._compute_transform(
               tuple(self.position),
               tuple(map(tuple, self.rotation)),
               self.scale)

    def update_transform(self, parent_transform=None):
        """Update the model matrix of this node and all of its descendants."""
        if parent_transform is None:
            self.model_matrix = self.get_local_transform()
        else:
            self.model_matrix = parent_transform.astype('f4') @ self.get_local_transform()

        for n in self.nodes:
            n.update_transform(self.model_matrix)

    @property
    def color(self):
        return self.material.color

    @color.setter
    def color(self, color):
        self.material.color = color

    @property
    def bounds(self):
        """ The bounds in the format ((x_min, x_max), (y_min, y_max), (z_min, z_max)) """
        return np.array([[0, 0], [0, 0], [0, 0]])

    @property
    def current_bounds(self):
        return np.array([[0, 0], [0, 0], [0, 0]])

    @property
    def current_center(self):
        return self.current_bounds.mean(-1)

    @property
    def center(self):
        return self.bounds.mean(-1)

    def get_local_bounds(self, points):
        if len(points.shape) == 2 and points.shape[-1] == 3:
            points = points[np.newaxis]
        assert len(points.shape) == 3

        # Compute min and max coordinates of the bounding box ignoring NaNs.
        val = np.array([
            [np.nanmin(points[:, :, 0]), np.nanmax(points[:, :, 0])],
            [np.nanmin(points[:, :, 1]), np.nanmax(points[:, :, 1])],
            [np.nanmin(points[:, :, 2]), np.nanmax(points[:, :, 2])]])

        # If any of the elements is NaN return an empty bounding box.
        if np.isnan(val).any():
            return np.array([[0, 0], [0, 0], [0, 0]])
        else:
            return val

    def get_bounds(self, points):
        val = self.get_local_bounds(points)

        # Transform bounding box with the model matrix.
        val = (self.model_matrix @ np.vstack((val, np.array([1.0, 1.0]))))[:3]

        # If any of the elements is NaN return an empty bounding box.
        if np.isnan(val).any():
            return np.array([[0, 0], [0, 0], [0, 0]])
        else:
            return val

    @property
    def n_frames(self):
        return self._n_frames

    @n_frames.setter
    def n_frames(self, n_frames):
        self._n_frames = n_frames

    def __len__(self):
        return self.n_frames

    @property
    def current_frame_id(self):
        return self._current_frame_id

    @current_frame_id.setter
    def current_frame_id(self, frame_id):
        # Check if the frame changed.
        last_frame_id = self._current_frame_id if self._enabled_frames is None else self._internal_frame_id
        if self.n_frames == 1 or frame_id == last_frame_id:
            return

        self.on_before_frame_update()
        if self._enabled_frames is None:
            if frame_id < 0:
                self._current_frame_id = 0
            elif frame_id >= len(self):
                self._current_frame_id = len(self) - 1
            else:
                self._current_frame_id = frame_id
        else:
            # If an enabled_frames is present use it to get the current frame.
            if frame_id < 0:
                self._internal_frame_id = 0
            elif frame_id >= self._enabled_frames.shape[0]:
                self._internal_frame_id = self._enabled_frames.shape[0] - 1
            else:
                self._internal_frame_id = frame_id
            self._current_frame_id = self._enabled_frame_id[self._internal_frame_id]
            # Update enabled using the mask.
            self.enabled = self._enabled_frames[self._internal_frame_id]

        # Update frame id of all children nodes.
        for n in self.nodes:
            n.current_frame_id = self._current_frame_id

        self.on_frame_update()
        if self.parent and (self._positions.shape[0] > 1 or self._rotations.shape[0] > 1 or self._scales.shape[0] > 1):
            self.update_transform(self.parent.model_matrix)

    def next_frame(self):
        self.current_frame_id = self.current_frame_id + 1 if self.current_frame_id < len(self) - 1 else 0

    def previous_frame(self):
        self.current_frame_id = self.current_frame_id - 1 if self.current_frame_id > 0 else len(self) - 1

    def on_before_frame_update(self):
        """Called when the current frame is about to change, 'self.current_frame_id' still has the id of the
           previous frame."""
        pass

    def on_frame_update(self):
        """Called when the current frame is changed."""
        pass

    def add(self, *nodes, **kwargs):
        self._add_nodes(*nodes, **kwargs)

    def _add_node(self,
                  n: 'Node',
                  show_in_hierarchy=True,
                  expanded=False,
                  enabled=True):
        """
        Add a single node
        :param show_in_hierarchy: Whether to show the node in the scene hierarchy.
        :param expanded: Whether the node is initially expanded in the GUI.
        """
        if n is None:
            return
        n._show_in_hierarchy = show_in_hierarchy
        n._expanded = expanded
        n._enabled = enabled if n._enabled_frames is None else n._enabled_frames[n.current_frame_id]
        self.nodes.append(n)
        n.parent = self
        n.update_transform(self.model_matrix)

    def _add_nodes(self, *nodes, **kwargs):
        """Add multiple nodes"""
        for n in nodes:
            self._add_node(n, **kwargs)

    def remove(self, *nodes):
        for n in nodes:
            n.release()
            self.nodes.remove(n)

    @property
    def show_in_hierarchy(self):
        return self._show_in_hierarchy

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        self._enabled = enabled

    @property
    def expanded(self):
        return self._expanded

    @expanded.setter
    def expanded(self, expanded):
        self._expanded = expanded

    def is_transparent(self):
        """
        Returns true if the object is transparent and should thus be sorted when rendering.
        Subclassess that use a different color should implement this method to be rendered correctly when transparent.
        """
        return self.material.color[3] < 1.0

    def gui(self, imgui):
        """
        Render GUI for custom node properties and controls. Implementation optional.
        Elements rendered here will show up in the scene hierarchy
        :param imgui: imgui context.
        See https://pyimgui.readthedocs.io/en/latest/reference/imgui.core.html for available elements to render
        """
        pass

    def gui_modes(self, imgui):
        """ Render GUI with toolbar (tools) for this particular node"""

    def gui_animation(self, imgui):
        """ Render GUI for animation related settings"""

        if self._enabled_frames is None:
            if self.n_frames > 1:
                u, fid = imgui.slider_int('Frame##r_{}'.format(self.unique_name),
                                        self.current_frame_id, min_value=0, max_value=self.n_frames - 1)
                if u:
                    self.current_frame_id = fid
        else:
            u, fid = imgui.slider_int('Frame##r_{}'.format(self.unique_name),
                                    self._internal_frame_id, min_value=0, max_value=self._enabled_frames.shape[0] - 1)
            if u:
                self.current_frame_id = fid

    def gui_affine(self, imgui):
        """ Render GUI for affine transformations"""
        # Position controls
        up, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 1e-2, format='%.2f')
        if up:
            self.position = pos

        # Rotation controls
        euler_angles = rot2euler_numpy(self.rotation[np.newaxis], degrees=True)[0]
        ur, euler_angles = imgui.drag_float3('Rotation##pos{}'.format(self.unique_name), *euler_angles, 1e-2, format='%.2f')
        if ur:
            self.rotation = euler2rot_numpy(np.array(euler_angles)[np.newaxis], degrees=True)[0]

        # Scale controls
        us, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.scale, 1e-2, min_value=0.001,
                                    max_value=100.0, format='%.3f')
        if us:
            self.scale = scale

    def gui_material(self, imgui):
        """ Render GUI with material properties """

        # Color Control
        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color, show_alpha=True)
        if uc:
            self.color = color

        # Diffuse
        ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name), self.material.diffuse,  0.0, 1.0, '%.2f')
        if ud:
            self.material.diffuse = diffuse

        # Ambient
        ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name), self.material.ambient,  0.0, 1.0, '%.2f')
        if ua:
            self.material.ambient = ambient

    def gui_io(self, imgui):
        """ Render GUI for import/export """
        pass

    def gui_mode_view(self, imgui):
        """ Render custom GUI for view mode """
        pass

    def gui_context_menu(self, imgui):
        _, self.enabled = imgui.checkbox("Enabled", self.enabled)
        if any([n._show_in_hierarchy for n in self.nodes]):
            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            for n in self.nodes:
                if not n._show_in_hierarchy:
                    continue
                if imgui.begin_menu(f"{n.name}##{n.uid}"):
                    n.gui_context_menu(imgui)
                    imgui.end_menu()

    # Renderable
    @staticmethod
    def once(func):
        def _decorator(self, *args, **kwargs):
            if self.is_renderable:
                return
            else:
                func(self, *args, **kwargs)
                self.is_renderable = True
        return _decorator

    def make_renderable(self, ctx):
        """
        Prepares this object for rendering. This function must be called before `render` is used.
        :param ctx: The moderngl context.
        """
        pass

    def render(self, camera, position=None, rotation=None, **kwargs):
        """Render the current frame in this sequence."""
        pass

    def render_positions(self, prog):
        """
        Render with a VAO with only positions bound, used for shadow mapping, fragmap and depth prepass.
        """
        pass

    def redraw(self, **kwargs):
        """ Perform update and redraw operations. Push to the GPU when finished. Recursively redraw child nodes"""
        for n in self.nodes:
            n.redraw(**kwargs)

    def set_camera_matrices(self, prog, camera, **kwargs):
        """Set the model view projection matrix in the given program."""
        # Transpose because np is row-major but OpenGL expects column-major.
        prog['model_matrix'].write(self.model_matrix.T.astype('f4').tobytes())
        prog['view_projection_matrix'].write(camera.get_view_projection_matrix().T.astype('f4').tobytes())

    def receive_shadow(self, program, **kwargs):
        """
        Call this function if the renderable is to receive shadows.
        :param program: The shader program that can shade with shadows.
        :param kwargs: The render kwargs.
        """
        if kwargs.get('shadows_enabled', False):
            lights = kwargs['lights']

            for i, light in enumerate(lights):
                if light.shadow_enabled and light.shadow_map:
                    light_matrix = light.mvp() @ self.model_matrix
                    program[f'dirLights[{i}].matrix'].write(light_matrix.T.tobytes())

                    # Bind shadowmap to slot i + 1, we reserve slot 0 for the mesh texture
                    # and use slots 1 to (#lights + 1) for shadow maps
                    light.shadow_map.use(location=i + 1)

            # Set sampler uniforms
            uniform = program[f'shadow_maps']
            uniform.value = 1 if uniform.array_length == 1 else [*range(1, len(lights) + 1)]

    def render_shadowmap(self, light_matrix):
        if not self.cast_shadow or self.depth_only_program is None or self.color[3] == 0.0:
            return

        prog = self.depth_only_program
        prog['model_matrix'].write(self.model_matrix.T.tobytes())
        prog['view_projection_matrix'].write(light_matrix.T.tobytes())

        self.render_positions(prog)

    def render_fragmap(self, ctx, camera, uid=None):
        if not self.fragmap or self.fragmap_program is None:
            return

        # Transpose because np is row-major but OpenGL expects column-major.
        prog = self.fragmap_program
        self.set_camera_matrices(prog, camera)

        # Render with the specified object uid, if None use the node uid instead.
        prog['obj_id'] = uid or self.uid

        if self.backface_culling or self.backface_fragmap:
            ctx.enable(moderngl.CULL_FACE)
        else:
            ctx.disable(moderngl.CULL_FACE)

        # If backface_fragmap is enabled for this node only render backfaces
        if self.backface_fragmap:
            ctx.cull_face = 'front'

        self.render_positions(prog)

        # Restore cull face to back
        if self.backface_fragmap:
            ctx.cull_face = 'back'

    def render_depth_prepass(self, camera, **kwargs):
        if not self.depth_prepass or self.depth_only_program is None:
            return

        prog = self.depth_only_program
        self.set_camera_matrices(prog, camera)
        self.render_positions(prog)

    def render_outline(self, ctx, camera):
        if self.outline and self.outline_program is not None:
            prog  = self.outline_program
            self.set_camera_matrices(prog, camera)

            if self.backface_culling:
                ctx.enable(moderngl.CULL_FACE)
            else:
                ctx.disable(moderngl.CULL_FACE)
            self.render_positions(prog)

        # Render children node recursively.
        for n in self.nodes:
            n.render_outline(ctx, camera)

    def release(self):
        """
        Release all OpenGL resources used by this node and any of its children. Subclasses that instantiate OpenGL
        objects should implement this method with '@hooked' to avoid leaking resources.
        """
        for n in self.nodes:
            n.release()

    def on_selection(self, node, instance_id, tri_id):
        """
        Called when the node is selected

        :param node:  the node which was clicked (can be None if the selection wasn't a mouse event)
        :param instance_id: the id of the instance that was clicked, 0 if the object is not instanced
                            (can be None if the selection wasn't a mouse event)
        :param tri_id: the id of the triangle that was clicked from the 'node' mesh
                       (can be None if the selection wasn't a mouse event)
        """
        pass

    def key_event(self, key, wnd_keys):
        """
        Handle shortcut key presses (if you are the selected object)
        """
        pass



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
import os
import trimesh
import tqdm
import re
import pickle

from aitviewer.scene.node import Node
from aitviewer.shaders import get_depth_only_program, get_flat_lit_with_edges_face_color_program, get_fragmap_program, get_outline_program
from aitviewer.shaders import get_smooth_lit_with_edges_face_color_program
from aitviewer.shaders import get_smooth_lit_with_edges_program
from aitviewer.shaders import get_flat_lit_with_edges_program
from aitviewer.shaders import get_smooth_lit_texturized_program
from aitviewer.utils import set_lights_in_program
from aitviewer.utils import set_material_properties
from aitviewer.utils.decorators import hooked
from aitviewer.utils.so3 import euler2rot_numpy, rot2euler_numpy
from aitviewer.utils.utils import compute_vertex_and_face_normals_sparse
from functools import lru_cache
from moderngl_window.opengl.vao import VAO
from PIL import Image
from trimesh.triangles import points_to_barycentric
import trimesh.geometry


class Meshes(Node):
    """A sequence of triangle meshes. This assumes that the mesh topology is fixed over the sequence."""

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 path_to_texture=None,
                 cast_shadow=True,
                 pickable=True,
                 flat_shading=False,
                 draw_edges=False,
                 draw_outline=False,
                 instance_transforms=None,
                 icon="\u008d",
                 **kwargs):
        """
        Initializer.
        :param vertices: A np array of shape (N, V, 3) or (V, 3).
        :param faces: A np array of shape (F, 3).
        :param vertex_normals: A np array of shape (N, V, 3). If not provided, the vertex normals will be computed,
          which incurs some overhead.
        :param face_normals: A np array of shape (N, F, 3). If not provided, the face normals will be computed, which
          incurs some overhead.
        :param vertex_colors: A np array of shape (N, V, 4) overriding the uniform color.
        :param face_colors: A np array of shape (N, F, 4) overriding the uniform or vertex colors.
        :param uv_coords: A np array of shape (V, 2) if the mesh is to be textured.
        :param path_to_texture: Path to an image file that serves as the texture.
        :param cast_shadow: If True the mesh casts a shadow on other objects.
        :param pickable: If True the mesh can be selected with a mouse click.
        :param flat_shading: If True the each face of the mesh is shaded with a constant normal.
        :param draw_edges: If True the normals the edges of the mesh is drawn on top of the mesh.
        :param draw_outline: If true an outline is drawn around the mesh.
        :instance_transforms: np array of size (N, I, 4, 4) or (I, 4, 4) or None. If not None, 'I' instances of
            the same mesh will be rendered, each with its own transformation matrix.
        """
        if len(vertices.shape) == 2 and vertices.shape[-1] == 3:
            vertices = vertices[np.newaxis]
        assert len(vertices.shape) == 3
        assert len(faces.shape) == 2
        n_frames = vertices.shape[0]

        # Instancing.
        if instance_transforms is not None:
            # Check shape of transforms.
            if len(instance_transforms.shape) == 3:
                instance_transforms = instance_transforms[np.newaxis]
            assert len(instance_transforms.shape) == 4

            # Number of instance frames must match number of frames or be 1.
            assert n_frames == 1 or instance_transforms.shape[0] == 1 or n_frames == instance_transforms.shape[0]
            n_frames = max(n_frames, instance_transforms.shape[0])

            self._instance_transforms = instance_transforms
        else:
            self._instance_transforms = None

        super(Meshes, self).__init__(n_frames=n_frames, icon=icon, **kwargs)

        self._vertices = vertices
        self._faces = faces.astype(np.int32)

        def _maybe_unsqueeze(x):
            return x[np.newaxis] if x is not None and x.ndim == 2 else x

        self._vertex_normals = _maybe_unsqueeze(vertex_normals)
        self._face_normals = _maybe_unsqueeze(face_normals)
        self.vertex_colors = _maybe_unsqueeze(vertex_colors)
        self.face_colors = _maybe_unsqueeze(face_colors)

        # Texture handling.
        self.has_texture = uv_coords is not None
        self.uv_coords = uv_coords

        if self.has_texture:
            self.use_pickle_texture = path_to_texture.endswith((".pickle", "pkl"))
            if self.use_pickle_texture:
                self.texture_image = pickle.load(open(path_to_texture, "rb"))
            else:
                self.texture_image = Image.open(path_to_texture).transpose(method=Image.FLIP_TOP_BOTTOM).convert("RGB")
        else:
            self.texture_image = None

        # Enable rendering passes
        self.cast_shadow = cast_shadow
        self.fragmap = pickable
        self.depth_prepass = True
        self.outline = True

        # Misc.
        self._flat_shading = flat_shading
        self.draw_edges = draw_edges
        self.draw_outline = draw_outline
        self.show_texture = self.has_texture
        self.norm_coloring = False
        self.normals_r = None
        self.need_upload = True
        self._use_uniform_color = self._vertex_colors is None and self._face_colors is None
        self._vertex_faces_sparse = trimesh.geometry.index_sparse(self._vertices.shape[1], self._faces)

    @classmethod
    def instanced(cls, *args, positions=None, rotations=None, scales=None, **kwargs):
        """
        Creates and returns an instanced sequence of N frames and I instances.
        Each instance will have its own position, rotation and scale.
        :param positions: np array of size (N, I, 3) or (I, 3) or None.
        :param rotations: np array of size (N, I, 3, 3) or (I, 3, 3) or None.
        :param scales: np array of size (N, I) or (I) or None.

        *args, and **kwargs are forwarded to the Meshes constructor.
        """
        assert positions is not None or rotations is not None or scales is not None

        n_instances = 0
        n_frames = 0
        def check_array(a, dim):
            nonlocal n_instances, n_frames
            if a is not None:
                if len(a.shape) == dim + 1:
                    a = a[np.newaxis]
                n_frames = max(n_frames, a.shape[0])
                n_instances = max(n_instances, a.shape[1])
            return a

        positions = check_array(positions, 1)
        rotations = check_array(rotations, 2)
        scales = check_array(scales, 0)

        if positions is None:
            positions = np.zeros((n_frames, n_instances, 3))
        if rotations is None:
            rotations = np.zeros((n_frames, n_instances, 3, 3))
            rotations[:, :] = np.eye(3)
        if scales is None:
            scales = np.ones((n_frames, n_instances))

        transforms = np.zeros((n_frames, n_instances, 4, 4))
        transforms[:, :, :3, :3] = (rotations.reshape((-1, 9)) * scales.reshape((-1, 1))).reshape((n_frames, n_instances, 3, 3))
        transforms[:, :, :3, 3] = positions
        transforms[:, :, 3, 3] = 1.0
        return cls(*args, **kwargs, instance_transforms=transforms)

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        if len(vertices.shape) == 2:
            vertices = vertices[np.newaxis]

        # Update vertices and redraw
        self._vertices = vertices
        self.n_frames = len(vertices)

        # If vertex or face normals were supplied, they are no longer valid.
        self._vertex_normals = None
        self._face_normals = None

        # Must clear all LRU caches where the vertices are used.
        self.compute_vertex_and_face_normals.cache_clear()

        self.redraw()

    @property
    def faces(self):
        return self._faces

    @faces.setter
    def faces(self, f):
        self._faces = f.astype(np.int32)
        self._vertex_faces_sparse = trimesh.geometry.index_sparse(self.vertices.shape[1], self._faces)

    @property
    def current_vertices(self):
        idx = self.current_frame_id if self.vertices.shape[0] > 1 else 0
        return self.vertices[idx]

    @current_vertices.setter
    def current_vertices(self, vertices):
        idx = self.current_frame_id if self.vertices.shape[0] > 1 else 0
        self._vertices[idx] = vertices
        self.compute_vertex_and_face_normals.cache_clear()
        self.redraw()

    @property
    def n_faces(self):
        return self.faces.shape[0]

    @property
    def n_vertices(self):
        return self.vertices.shape[1]

    @property
    def vertex_faces(self):
        # To compute the normals we need to know a mapping from vertex ID to all faces that this vertex is part of.
        # Because we are lazy we abuse trimesh to compute this for us. Not all vertices have the maximum degree, so
        # this array is padded with -1 if necessary.
        return trimesh.Trimesh(self.vertices[0], self.faces, process=False).vertex_faces

    @property
    def vertex_normals(self):
        """Get or compute all vertex normals (this might take a while for long sequences)."""
        if self._vertex_normals is None:
            vertex_normals, _ = compute_vertex_and_face_normals_sparse(self.vertices, self.faces, self._vertex_faces_sparse,
                                                                normalize=True)
            self._vertex_normals = vertex_normals
        return self._vertex_normals

    @property
    def face_normals(self):
        """Get or compute all face normals (this might take a while for long sequences)."""
        if self._face_normals is None:
            _, face_normals = compute_vertex_and_face_normals_sparse(self.vertices, self.faces, self._vertex_faces_sparse,
                                                              normalize=True)
            self._face_normals = face_normals
        return self._face_normals

    def vertex_normals_at(self, frame_id):
        """Get or compute the vertex normals at the given frame."""
        if self._vertex_normals is None:
            vn, _ = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._vertex_normals.shape) == 3, f"Got shape {self._vertex_normals.shape}"
            vn = self._vertex_normals[frame_id]
        return vn

    def face_normals_at(self, frame_id):
        """Get or compute the face normals at the given frame."""
        if self._face_normals is None:
            _, fn = self.compute_vertex_and_face_normals(frame_id, normalize=True)
        else:
            assert len(self._face_normals.shape) == 3, f"Got shape {self._face_normals.shape}"
            fn = self._face_normals[frame_id]
        return fn

    @property
    def vertex_colors(self):
        if self._vertex_colors is None:
            self._vertex_colors = np.full((self.n_frames, self.n_vertices, 4), self.material.color)
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, vertex_colors):
        # If vertex_colors are None, we resort to the material color.
        if vertex_colors is None:
            self._vertex_colors = None
            self._use_uniform_color = True
        elif isinstance(vertex_colors, tuple) and len(vertex_colors) == 4:
            self.vertex_colors = None
            self._use_uniform_color = True
            self.material.color = vertex_colors
        else:
            if len(vertex_colors.shape) == 2:
                assert vertex_colors.shape[0] == self.n_vertices
                vertex_colors = np.repeat(vertex_colors[np.newaxis], self.n_frames, axis=0)
            assert len(vertex_colors.shape) == 3
            self._vertex_colors = vertex_colors
            self._use_uniform_color = False
            self.redraw()

    @property
    def current_vertex_colors(self):
        if self._use_uniform_color:
            return np.full((self.n_vertices, 4), self.material.color)
        else:
            idx = self.current_frame_id if self.vertex_colors.shape[0] > 1 else 0
            return self.vertex_colors[idx]

    @property
    def face_colors(self):
        return self._face_colors

    @face_colors.setter
    def face_colors(self, face_colors):
        if face_colors is not None:
            if len(face_colors.shape) == 2:
                face_colors = face_colors[np.newaxis]
            self._face_colors = face_colors
            self._use_uniform_color = False
        else:
            self._face_colors = None
        self.redraw()

    @property
    def current_face_colors(self):
        if self._use_uniform_color:
            return np.full((self.n_faces, 4), self.material.color)
        else:
            idx = self.current_frame_id if self.face_colors.shape[0] > 1 else 0
            return self.face_colors[idx]

    @Node.color.setter
    def color(self, color):
        self.material.color = color

        if self.face_colors is None:
            self.vertex_colors = color

    @property
    def flat_shading(self):
        return self._flat_shading

    @flat_shading.setter
    def flat_shading(self, flat_shading):
        if self._flat_shading != flat_shading:
            self._flat_shading = flat_shading
            self.redraw()

    def closest_vertex_in_triangle(self, tri_id, point):
        face_vertex_id = np.linalg.norm((self.current_vertices[self.faces[tri_id]] - point), axis=-1).argmin()
        return self.faces[tri_id][face_vertex_id]

    def get_bc_coords_from_points(self, tri_id, points):
        return points_to_barycentric(self.current_vertices[self.faces[[tri_id]]], points)[0]

    @lru_cache(2048)
    def compute_vertex_and_face_normals(self, frame_id, normalize=False):
        """
        Compute face and vertex normals for the given frame. We use an LRU cache since this is a potentially
        expensive operation. This function exists because computing the normals on all frames can increase the
        startup time of the viewer considerably.

        :param frame_id: On which frame to compute the normals.
        :param normalize: Whether or not to normalize the normals. Not doing it is faster and the shaders typically
          enforce unit length of normals anyway.
        :return: The vertex and face normals as a np arrays of shape (V, 3) and (F, 3) respectively.
        """
        vs = self.vertices[frame_id:frame_id + 1] if self.vertices.shape[0] > 1 else self.vertices
        vn, fn = compute_vertex_and_face_normals_sparse(vs, self.faces, self._vertex_faces_sparse, normalize)
        return vn.squeeze(0), fn.squeeze(0)

    @property
    def bounds(self):
        if self.instance_transforms is None:
            return self.get_bounds(self.vertices)
        else:
            # Get bounds in local coordinates
            bounds = self.get_local_bounds(self.vertices)

            # Transform bounds with instance transforms
            min = np.append(bounds[:, 0], 1.0)
            max = np.append(bounds[:, 1], 1.0)
            transforms = self.instance_transforms.reshape((-1, 4, 4))
            mins = transforms @ min
            maxs = transforms @ max

            # Return bounds in world coordinates
            return self.get_bounds(np.vstack((mins, maxs)))

    @property
    def current_bounds(self):
        if self.instance_transforms is None:
            return self.get_bounds(self.current_vertices)
        else:
            # Get bounds in local coordinates
            bounds = self.get_local_bounds(self.current_vertices)

            # Transform bounds with instance transforms
            min = np.append(bounds[:, 0], 1.0)
            max = np.append(bounds[:, 1], 1.0)
            transforms = self.current_instance_transforms.reshape((-1, 4, 4))
            mins = transforms @ min
            maxs = transforms @ max

            # Return bounds in world coordinates
            return self.get_bounds(np.vstack((mins[:, :3], maxs[:, :3])))

    def is_transparent(self):
        return self.color[3] < 1.0

    def on_frame_update(self):
        """Called whenever a new frame must be displayed."""
        super().on_frame_update()
        self.redraw()

    @property
    def current_instance_transforms(self):
        if self._instance_transforms is None:
            return None
        idx = self.current_frame_id if self._instance_transforms.shape[0] > 1 else 0
        return self._instance_transforms[idx]

    @property
    def instance_transforms(self):
        return self._instance_transforms

    @instance_transforms.setter
    def instance_transforms(self, instance_transforms):
        assert self._instance_transforms.shape == instance_transforms
        self._instance_transforms = instance_transforms

    @property
    def n_instances(self):
        if self._instance_transforms is None:
            return 1
        else:
            return self._instance_transforms.shape[1]

    def _upload_buffers(self):
        """Upload the current frame data to the GPU for rendering."""
        if not self.is_renderable or not self._need_upload:
            return

        self._need_upload = False

        # Write positions.
        self.vbo_vertices.write(self.current_vertices.astype('f4').tobytes())

        # Write normals.
        if not self.flat_shading:
            vertex_normals = self.vertex_normals_at(self.current_frame_id)
            self.vbo_normals.write(vertex_normals.astype('f4').tobytes())

        if self.face_colors is None:
            # Write vertex colors.
            self.vbo_colors.write(self.current_vertex_colors.astype('f4').tobytes())
        else:
            # Write face colors.
            self.face_colors_texture.write(self.current_face_colors.astype('f4').tobytes())

        # Write uvs.
        if self.has_texture:
            self.vbo_uvs.write(self.uv_coords.astype('f4').tobytes())

        # Write instance transforms.
        if self.instance_transforms is not None:
            self.vbo_instance_transforms.write(np.transpose(self.current_instance_transforms.astype('f4'), (0, 2, 1)).tobytes())

    def redraw(self, **kwargs):
        self._need_upload = True

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx: moderngl.Context):
        """Prepares this object for rendering. This function must be called before `render` is used."""

        vs = "lit_with_edges.glsl"
        positions_vs = "mesh_positions.vs.glsl"
        instanced = 1 if self.instance_transforms is not None else 0
        self.smooth_prog = get_smooth_lit_with_edges_program(vs, instanced)
        self.flat_prog = get_flat_lit_with_edges_program(vs, instanced)
        self.smooth_face_prog = get_smooth_lit_with_edges_face_color_program(vs, instanced)
        self.flat_face_prog = get_flat_lit_with_edges_face_color_program(vs, instanced)

        self.depth_only_program = get_depth_only_program(positions_vs, instanced)
        self.outline_program = get_outline_program(positions_vs, instanced)
        self.fragmap_program = get_fragmap_program(positions_vs, instanced)

        vertices = self.current_vertices
        vertex_normals = self.vertex_normals_at(self.current_frame_id)
        vertex_colors = self.current_vertex_colors

        self.vbo_vertices = ctx.buffer(vertices.astype('f4').tobytes())
        self.vbo_normals = ctx.buffer(vertex_normals.astype('f4').tobytes())
        self.vbo_colors = ctx.buffer(vertex_colors.astype('f4').tobytes())
        self.vbo_indices = ctx.buffer(self.faces.tobytes())

        self.vao = VAO()
        self.vao.buffer(self.vbo_vertices, '3f4', 'in_position')
        self.vao.buffer(self.vbo_normals, '3f4', 'in_normal')
        self.vao.buffer(self.vbo_colors, '4f4', 'in_color')
        self.vao.index_buffer(self.vbo_indices)

        if self.instance_transforms is not None:
            self.vbo_instance_transforms = ctx.buffer(np.transpose(self.current_instance_transforms.astype('f4'), (0, 2, 1)).tobytes())
            self.vao.buffer(self.vbo_instance_transforms, '16f4/i', 'instance_transform')

        self.face_colors_texture = ctx.texture((self.faces.shape[0], 1), 4, dtype='f4')
        if self.face_colors is not None:
            self.face_colors_texture.write(self.current_face_colors.astype('f4').tobytes())

        if self.has_texture:
            img = self.texture_image
            if self.use_pickle_texture:
                self.texture = ctx.texture(img.shape[:2], img.shape[2], img.tobytes())
            else:
                self.texture = ctx.texture(img.size, 3, img.tobytes())
            self.texture_prog = get_smooth_lit_texturized_program(vs)
            self.vbo_uvs = ctx.buffer(self.uv_coords.astype('f4').tobytes())
            self.vao.buffer(self.vbo_uvs, '2f4', 'in_uv')

    @hooked
    def release(self):
        if self.is_renderable:
            self.vao.release()
            if self.has_texture:
                self.texture.release()

    def render(self, camera, **kwargs):
        self._upload_buffers()

        if self.has_texture and self.show_texture:
            prog = self.texture_prog
            prog['diffuse_texture'] = 0
            self.texture.use(0)
        else:
            if self.face_colors is None:
                if self.flat_shading:
                    prog = self.flat_prog
                else:
                    prog = self.smooth_prog
            else:
                if self.flat_shading:
                    prog = self.flat_face_prog
                else:
                    prog = self.smooth_face_prog
                self.face_colors_texture.use(0)
                prog["face_colors"] = 0
            prog['norm_coloring'].value = self.norm_coloring

        prog['use_uniform_color'] = self._use_uniform_color
        prog['uniform_color'] = self.material.color
        prog['draw_edges'].value = 1.0 if self.draw_edges else 0.0
        prog['win_size'].value = kwargs['window_size']

        self.set_camera_matrices(prog, camera, **kwargs)
        set_lights_in_program(prog, kwargs['lights'], kwargs['shadows_enabled'], kwargs['ambient_strength'])
        set_material_properties(prog, self.material)
        self.receive_shadow(prog, **kwargs)
        self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_instances)

    def render_positions(self, prog):
        if self.is_renderable:
            self._upload_buffers()
            self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_instances)

    def _show_normals(self):
        """Create and add normals at runtime"""
        vn = self.vertex_normals

        bounds = self.bounds
        diag = np.linalg.norm(bounds[:, 0] - bounds[:, 1])

        length = 0.005 * max(diag, 1) / self.scale
        vn = vn / np.linalg.norm(vn, axis=-1, keepdims=True) * length

        # Must import here because if we do it at the top we create a circular dependency.
        from aitviewer.renderables.arrows import Arrows

        positions = self.vertices
        self.normals_r = Arrows(positions, positions + vn,
                                r_base=length / 10, r_head=2 * length / 10, p=0.25, name='Normals')
        self.normals_r.current_frame_id = self.current_frame_id
        self.add(self.normals_r)

    def gui(self, imgui):
        super(Meshes, self).gui(imgui)

        _, self.show_texture = imgui.checkbox('Render Texture##render_texture{}'.format(self.unique_name),
                                              self.show_texture)
        _, self.norm_coloring = imgui.checkbox('Norm Coloring##norm_coloring{}'.format(self.unique_name),
                                               self.norm_coloring)
        _, self.flat_shading = imgui.checkbox('Flat shading [F]##flat_shading{}'.format(self.unique_name),
                                              self.flat_shading)
        _, self.draw_edges = imgui.checkbox('Draw edges [E]##draw_edges{}'.format(self.unique_name),
                                            self.draw_edges)
        _, self.draw_outline = imgui.checkbox('Draw outline##draw_outline{}'.format(self.unique_name),
                                              self.draw_outline)

        if self.normals_r is None:
            if imgui.button('Show Normals ##show_normals{}'.format(self.unique_name)):
                self._show_normals()

    def gui_context_menu(self, imgui):
        _, self.flat_shading = imgui.menu_item("Flat shading", "F", selected=self.flat_shading, enabled=True)
        _, self.draw_edges = imgui.menu_item("Draw edges", "E", selected=self.draw_edges, enabled=True)
        _, self.draw_outline = imgui.menu_item("Draw outline", selected=self.draw_outline)

        imgui.spacing()
        imgui.separator()
        imgui.spacing()
        super().gui_context_menu(imgui)


    def gui_io(self, imgui):
        if imgui.button('Export OBJ##export_{}'.format(self.unique_name)):
            mesh = trimesh.Trimesh(vertices=self.current_vertices, faces=self.faces, process=False)
            mesh.export('../export/' + self.name + '.obj')

    def key_event(self, key, wnd_keys):
        if key == wnd_keys.F:
            self.flat_shading = not self.flat_shading
        elif key == wnd_keys.E:
            self.draw_edges = not self.draw_edges


class VariableTopologyMeshes(Node):
    """
    A sequence of meshes that do not share the same topology (i.e. different nr of vertices, faces, texture, etc.).
    This simply treats every time instance as a separate mesh, so it's not optimized for performance.
    """

    def __init__(self,
                 vertices,
                 faces,
                 vertex_normals=None,
                 face_normals=None,
                 vertex_colors=None,
                 face_colors=None,
                 uv_coords=None,
                 texture_paths=None,
                 preload=True,
                 **kwargs):
        """
        Initializer.
        :param vertices: A list of length N with np arrays of size (V_n, 3).
        :param faces: A list of length N with np arrays of shape (F_n, 3).
        :param vertex_normals: An optional list of length N with np arrays of shape (V_n, 3).
        :param face_normals: An optional list of length N with np arrays of shape (F_n, 3).
        :param vertex_colors: An optional list of length N with np arrays of shape (V_n, 4) overriding the
          uniform color.
        :param face_colors: An optional list of length N with np arrays of shape (F_n, 4) overriding the uniform
          or vertex colors.
        :param uv_coords: An optional list of length N with np arrays of shape (V_n, 2) if the mesh is to be textured.
        :param texture_paths: An optional list of length N containing paths to the texture as an image file.
        :param preload: Whether or not to pre-load all the meshes. This increases loading time and memory consumption,
          but allows interactive animations.
        """
        assert len(vertices) == len(faces)
        super(VariableTopologyMeshes, self).__init__(n_frames=len(vertices))
        self.preload = preload

        self.vertices = vertices
        self.faces = faces
        self.vertex_normals = vertex_normals
        self.face_normals = face_normals
        self.vertex_colors = vertex_colors
        self.face_colors = face_colors
        self.uv_coords = uv_coords
        self.texture_paths = texture_paths
        self.mesh_kwargs = kwargs

        self._current_mesh = dict()  # maps from frame ID to mesh
        self._all_meshes = []
        if self.preload:
            for f in range(self.n_frames):
                m = self._construct_mesh_at_frame(f)
                self._all_meshes.append(m)
        # Set to true after changing the color and used when preloading
        # is not enabled to override the current mesh color with the new color
        self._override_color = False

        self.show_texture = True
        self.norm_coloring = False
        self.flat_shading = False
        self.draw_edges = False
        self.ctx = None

    def _construct_mesh_at_frame(self, frame_id):
        m = Meshes(self.vertices[frame_id],
                   self.faces[frame_id],
                   self.vertex_normals[frame_id] if self.vertex_normals is not None else None,
                   self.face_normals[frame_id] if self.face_normals is not None else None,
                   self.vertex_colors[frame_id] if self.vertex_colors is not None else None,
                   self.face_colors[frame_id] if self.face_colors is not None else None,
                   self.uv_coords[frame_id] if self.uv_coords is not None else None,
                   self.texture_paths[frame_id] if self.texture_paths is not None else None,
                   **self.mesh_kwargs)
        return m

    @classmethod
    def from_trimeshes(cls, trimeshes, **kwargs):
        """Initialize from a list of trimeshes."""
        vertices = []
        faces = []
        vertex_normals = []
        for m in trimeshes:
            vertices.append(m.vertices)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
        meshes = cls(vertices, faces, vertex_normals, **kwargs)
        return meshes

    @classmethod
    def from_plys(cls, plys, **kwargs):
        """Initialize from a list paths to .ply files."""
        vertices, faces, vertex_normals, vertex_colors = [], [], [], []
        sc = kwargs.get('vertex_scale', 1.0)
        for i in tqdm.tqdm(range(len(plys))):
            m = trimesh.load(plys[i])
            vertices.append(m.vertices * sc)
            faces.append(m.faces)
            vertex_normals.append(m.vertex_normals)
            vertex_colors.append(m.visual.vertex_colors / 255.0)
        return cls(vertices, faces, vertex_normals, vertex_colors=vertex_colors)

    @classmethod
    def from_directory(cls, path, preload=False, vertex_scale=1.0, high_quality=False, **kwargs):
        """
        Initialize from a directory containing mesh and texture data.

        Mesh files must be in pickle (.pkl) or obj (.obj) format, their name
        must start with 'mesh' and end with a frame number.

        Texture files must be in pickle (.pkl), png (.png) or jpeg (.jpg or .jpeg) format
        and their name must match the respective mesh filename with 'atlas' instead of 'mesh'
        at the start of the filename.

        Example:
        path/mesh_001.obj
        path/mesh_002.obj
        path/atlas_001.png
        path/atlas_002.png

        A mesh pickle file must be a dictionary of numpy array with the following shapes:
        {
            'vertices': (V, 3)
            'normals':  (V, 3)
            'uvs':      (V, 2)
            'faces':    (F, 3)
        }

        A texture pickle file must be a raw RGB image stored as a numpy array of shape (width, height, 3)
        """

        files = os.listdir(path)

        # Supported mesh formats in order of preference (fastest to slowest)
        mesh_supported_formats = [".pkl", ".obj"]

        mesh_format = None
        for format in mesh_supported_formats:
            if any(map(lambda x: x.startswith("mesh") and x.endswith(format), files)):
                mesh_format = format
                break

        if mesh_format is None:
            raise ValueError(
                f'Unable to find mesh with supported extensions ({", ".join(mesh_supported_formats)}) at {path}')

        # Supported texture formats in order of preference (fastest to slowest)
        texture_supported_formats = [".pkl", ".jpg", ".jpeg", ".png"]

        # If high_quality is set to true prioritize PNGs
        if high_quality:
            texture_supported_formats = [".png", ".jpg", ".jpeg", ".pkl"]

        texture_format = None
        for format in texture_supported_formats:
            if any(map(lambda x: x.startswith("atlas") and x.endswith(format), files)):
                texture_format = format
                break

        if texture_format is None:
            raise ValueError(
                f'Unable to find atlas with supported extensions ({", ".join(texture_supported_formats)}) at {path}')

        # Load all objects sorted by the keyframe number specified in the file name
        regex = re.compile(r"(\d*)$")

        def sort_key(x):
            name = os.path.splitext(x)[0]
            return int(regex.search(name).group(0))

        obj_names = filter(lambda x: x.startswith("mesh") and x.endswith(mesh_format), files)
        obj_names = sorted(obj_names, key=sort_key)

        vertices, faces, vertex_normals, uvs = [], [], [], []
        texture_paths = []

        for obj_name in tqdm.tqdm(obj_names):
            if mesh_format == ".pkl":
                mesh = pickle.load(open(os.path.join(path, obj_name), "rb"))
                vertices.append(mesh['vertices'] * vertex_scale)
                faces.append(mesh['faces'])
                vertex_normals.append(mesh['normals'])
                uvs.append(mesh['uvs'])
            else:
                mesh = trimesh.load(os.path.join(path, obj_name), process=False)
                vertices.append(mesh.vertices * vertex_scale)
                faces.append(mesh.faces)
                vertex_normals.append(mesh.vertex_normals)
                uvs.append(np.array(mesh.visual.uv).squeeze())

            texture_paths.append(
                os.path.join(path, obj_name.replace("mesh", "atlas").replace(mesh_format, texture_format)))

        return cls(vertices, faces, vertex_normals, uv_coords=uvs,
                   texture_paths=texture_paths, preload=preload, **kwargs)

    @property
    def current_mesh(self):
        if self.preload:
            return self._all_meshes[self.current_frame_id]
        else:
            m = self._current_mesh.get(self.current_frame_id, None)
            if m is None:
                # Need to construct a new one and clean up the old one.
                for k in self._current_mesh:
                    self._current_mesh[k].release()
                m = self._construct_mesh_at_frame(self.current_frame_id)

                # Set mesh position and scale
                m.update_transform(self.model_matrix)

                # Set mesh material
                m.material = self.material

                # Set draw settings.
                m.flat_shading = self.flat_shading
                m.draw_edges = self.draw_edges

                # Override the mesh color only if it has been changed by the user
                if self._override_color:
                    m.color = self.color

                m.make_renderable(self.ctx)
                self._current_mesh = {self.current_frame_id: m}
            return m

    @property
    def bounds(self):
        # TODO: this currently only returns the current mesh bounds for performance/simplicity.
        return self.current_mesh.bounds

    @property
    def current_bounds(self):
        return self.current_mesh.current_bounds

    def closest_vertex_in_triangle(self, tri_id, point):
        return self.current_mesh.closest_vertex_in_triangle(tri_id, point)

    def get_bc_coords_from_points(self, tri_id, points):
        return self.current_mesh.get_bc_coords_from_points(tri_id, points)

    def is_transparent(self):
        return self.color[3] < 1.0

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        if self.preload:
            for m in self._all_meshes:
                m.make_renderable(ctx)
        else:
            # A bit hacky so that we can make dynamically loaded meshes renderable when we load them.
            self.ctx = ctx
            self.current_mesh.make_renderable(ctx)

    def render(self, camera, **kwargs):
        self.current_mesh.show_texture = self.show_texture
        self.current_mesh.norm_coloring = self.norm_coloring
        self.current_mesh.flat_shading = self.flat_shading
        self.current_mesh.draw_edges = self.draw_edges
        self.current_mesh.render(camera, **kwargs)

    def render_depth_prepass(self, camera, **kwargs):
        self.current_mesh.render_depth_prepass(camera, **kwargs)

    def render_shadowmap(self, light_mvp, program):
        self.current_mesh.render_shadowmap(light_mvp, program)

    def render_fragmap(self, ctx, camera, prog, uid=None):
        # Since the current mesh is not a child node we cannot capture its selection.
        # Therefore we draw to the fragmap using our own id instead of the mesh id.
        self.current_mesh.render_fragmap(ctx, camera, prog, self.uid)

    def render_outline(self, *args, **kwargs):
        self.current_mesh.render_outline(*args, **kwargs)

    def gui_context_menu(self, imgui):
        _, self.flat_shading = imgui.menu_item("Flat shading", "F", selected=self.flat_shading, enabled=True)
        _, self.draw_edges = imgui.menu_item("Draw edges", "E", selected=self.draw_edges, enabled=True)
        _, self.draw_outline = imgui.menu_item("Draw outline", selected=self.draw_outline)

        imgui.spacing()
        imgui.separator()
        imgui.spacing()
        super().gui_context_menu(imgui)

    def gui_affine(self, imgui):
        """ Render GUI for affine transformations"""
        # Position controls
        up, pos = imgui.drag_float3('Position##pos{}'.format(self.unique_name), *self.position, 1e-2, format='%.2f')
        if up:
            self.position = pos

        # Rotation controls
        euler_angles = rot2euler_numpy(self.rotation[np.newaxis], degrees=True)[0]
        ur, euler_angles = imgui.drag_float3('Rotation##pos{}'.format(self.unique_name), *euler_angles, 1e-2,
                                             format='%.2f')
        if ur:
            self.rotation = euler2rot_numpy(np.array(euler_angles)[np.newaxis], degrees=True)[0]

        # Scale controls
        us, scale = imgui.drag_float('Scale##scale{}'.format(self.unique_name), self.scale, 1e-2, min_value=0.001,
                                     max_value=100.0, format='%.3f')
        if us:
            self.scale = scale

        if up or ur or us:
            if self.preload:
                for m in self._all_meshes:
                    m.update_transform(self.model_matrix)
            else:
                self.current_mesh.update_transform((self.model_matrix))

    def gui_material(self, imgui, show_advanced=True):
        # Color Control
        uc, color = imgui.color_edit4("Color##color{}'".format(self.unique_name), *self.material.color, show_alpha=True)
        if uc:
            self.color = color
            # If meshes are already loaded go through all and update the vertex colors
            if self.preload:
                for m in self._all_meshes:
                    m.color = color
            # Otherwise only update the current mesh and enable color override for other meshes
            else:
                self.current_mesh.color = color
                self._override_color = True

        _, self.show_texture = imgui.checkbox('Render Texture', self.show_texture)
        _, self.norm_coloring = imgui.checkbox('Norm Coloring', self.norm_coloring)
        _, self.flat_shading = imgui.checkbox('Flat shading [F]', self.flat_shading)
        _, self.draw_edges = imgui.checkbox('Draw edges [E]', self.draw_edges)
        _, self.draw_outline = imgui.checkbox('Draw outline', self.draw_outline)

        if show_advanced:
            if imgui.tree_node("Advanced material##advanced_material{}'".format(self.unique_name)):
                # Diffuse
                ud, diffuse = imgui.slider_float('Diffuse##diffuse{}'.format(self.unique_name),
                                                 self.current_mesh.material.diffuse,
                                                 0.0, 1.0, '%.2f')

                if ud:
                    self.material.diffuse = diffuse
                    # If meshes are already loaded go through all and update the diffuse value
                    if self.preload:
                        for m in self._all_meshes:
                            m.material.diffuse = diffuse
                    # Otherwise only update the current mesh, other meshes will be updated when loaded
                    else:
                        self.current_mesh.material.diffuse = diffuse

                # Ambient
                ua, ambient = imgui.slider_float('Ambient##ambient{}'.format(self.unique_name),
                                                 self.current_mesh.material.ambient,
                                                 0.0, 1.0, '%.2f')
                if ua:
                    self.material.ambient = ambient
                    # If meshes are already loaded go through all and update the ambient value
                    if self.preload:
                        for m in self._all_meshes:
                            m.material.ambient = ambient
                    # Otherwise only update the current mesh, other meshes will be updated when loaded
                    else:
                        self.current_mesh.material.ambient = ambient

                imgui.tree_pop()

    def key_event(self, key, wnd_keys):
        if key == wnd_keys.F:
            self.flat_shading = not self.flat_shading
        elif key == wnd_keys.E:
            self.draw_edges = not self.draw_edges

    @hooked
    def release(self):
        if self.preload:
            for m in self._all_meshes:
                m.release()
        else:
            m = self._current_mesh.get(self.current_frame_id, None)
            if m is not None:
                m.release()
